'use strict';

exports.__esModule = true;
exports['default'] = flatten;
var notEmpty = function notEmpty(x) {
  return x !== '' && x != null;
};

var isArray = function isArray(x) {
  return Array.isArray(x);
};
var isObject = function isObject(x) {
  return x.constructor === Object;
};
var isString = function isString(x) {
  return x.constructor === String;
};

var mergeArrays = function mergeArrays(a, b) {
  return (a.push.apply(a, b), a);
};
var mergeObjects = function mergeObjects(a, b) {
  return Object.assign(a, b);
};
var concatenate = function concatenate(a, b) {
  return a + b;
};

var merge = function merge(a, b) {
  if (a.constructor === b.constructor) {
    if (isArray(a)) return mergeArrays(a, b);
    if (isObject(a)) return mergeObjects(a, b);
    return a + b;
  }

  if (isString(a)) return b;
  if (isString(b)) return a;
  if (isArray(a)) return (a.push(b), a);
  if (isArray(b)) return (b.push(a), b);

  throw new Error('Unexpected case');
};

function flattenSequence(values) {
  return values.filter(notEmpty).reduce(merge, '');
}

function flattenRepetition(values, named) {
  if (named && values.length === 0) {
    return [];
  }

  if (values.some(isObject)) {
    return values.filter(isObject);
  }

  if (values.some(isArray)) {
    return values.filter(isArray).reduce(mergeArrays, []);
  }

  return values.reduce(concatenate, '');
}

function flatten(values) {
  var named = arguments[1] === undefined ? false : arguments[1];

  if (!isArray(values)) {
    return values;
  }

  var tag = values[0];
  var tail = values.slice(1);

  var result = tail.map(function (x) {
    return flatten(x);
  });

  switch (tag) {
    case 'maybe':
      return named ? result[0] : result[0] || '';
    case 'repetition':
      return flattenRepetition(result, named);
    case 'sequence':
      return flattenSequence(result);
  }

  throw new Error('Unknown tag \'' + tag + '\'');
}

module.exports = exports['default'];